'''
Enables fast on-demand sampling from a categorical probability distribution
'''

import numpy.random as r
import numpy as np


class Categorical(object):

	def __init__(self, scores):
		if len(scores) < 1:
			raise ValueError('The scores list must have length >= 1')
		self.scores = scores
		self.total = float(sum(scores))
		self.K = len(scores)
		self.setup()


	def get_probability(self, k):
		'''
		Get the actual probability associated to outcome k
		'''
		return self.orig_prob_mass[k]


	def setup(self):
		self.orig_prob_mass = np.zeros(self.K)
		self.mixture_prob_mass = np.zeros(self.K)
		self.mass_reasignments = np.zeros(self.K, dtype=np.int64)
 
		# Sort the data into the outcomes with probabilities
		# that are larger and smaller than 1/K.
		smaller = []
		larger  = []
		for k, score in enumerate(self.scores):
			self.orig_prob_mass[k] = score / self.total
			self.mixture_prob_mass[k] = (self.K*score) / self.total
			if self.mixture_prob_mass[k] < 1.0:
				smaller.append(k)
			else:
				larger.append(k)
 
		# We will have k different slots. Each slot represents 1/K
		# prbability mass, and to each we allocate all of the probability
		# mass from a "small" outcome, plus some probability mass from
		# a "large" outcome (enough to equal the total 1/K).
		# We keep track of the remaining mass left for the larger outcome,
		# allocating the remainder to another slot later.
		# The result is that the kth has some mass allocated to outcome
		# k, and some allocated to another outcome, recorded in J[k].
		# q[k] keeps track of how much mass belongs to outcome k, and 
		# how much belongs to outcome J[k].
		while len(smaller) > 0 and len(larger) > 0:
			small_idx = smaller.pop()
			large_idx = larger.pop()
 
			self.mass_reasignments[small_idx] = large_idx
			self.mixture_prob_mass[large_idx] = (
				self.mixture_prob_mass[large_idx] -
				(1.0 - self.mixture_prob_mass[small_idx])
			)
 
			if self.mixture_prob_mass[large_idx] < 1.0:
				smaller.append(large_idx)
			else:
				larger.append(large_idx)
 
		return self.mass_reasignments, self.mixture_prob_mass


	def sample(self, shape=()):

		if len(shape) < 1:
			return self._sample()
		else:
			this_dim = shape[0]
			recurse_shape = shape[1:]
			return np.array(
				[self.sample(recurse_shape) for i in range(this_dim)]
			, dtype='int64')


	def _sample(self):

		# Draw from the overall uniform mixture.
		k = np.int64(int(np.floor(r.rand()*self.K)))
	 
		# Draw from the binary mixture, either keeping the
		# small one, or choosing the associated larger one.
		if r.rand() < self.mixture_prob_mass[k]:
			return k
		else:
			return self.mass_reasignments[k]
 




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
#with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#    long_description = f.read()

setup(
    name='categorical',

    # Versions should comply with PEP440.  For a discussion on 
	# single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.3',

    description='Fast on-demand sampling from categorical distributions',
    long_description='''
Categorical Sampler
-----

Install from pip: `pip install categorical-sampler`

Let&rsquo;s generate a probability distribution to get us started.  First, sample a bunch of random numbers to determine probability &ldquo;scores&rdquo;.


    >>> from random import random
    >>> k = 10**6
    >>> scores = [random() for i in range(k)]
    >>> total = sum(scores)
    >>> probabilities = [s / total for s in scores]


We've normalized the scores to sum to 1, i.e. make
them into proper probabilities, but actually the categorical sampler will do that for us, so it&rsquo;s not necessary:

    >>> from categorical import Categorical as C
    >>> my_sampler = C(scores)
    >>> print my_sampler.sample()
    487702

Comparing to numpy, assuming we draw 1000 individual samples *individually*:


    >>> from numpy.random import choice
    >>> import time
    >>> 
    >>> def time_numpy():
    >>>     start = time.time()
    >>>     for i in range(1000):
    >>>         choice(k, p=probabilities)
    >>>     print time.time() - start
    >>> 
    >>> def time_my_alias():
    >>>     start = time.time()
    >>>     for i in range(1000):
    >>>         my_sampler.sample()
    >>>     print time.time() - start
    >>> 
    >>> time_numpy()
    31.0555009842
    >>> time_my_alias()
    0.0127031803131

Get the actual probability of a given outcome:

    >>> my_sampler.get_probability(487702)
    1.0911282101090306e-06 

	''',

    # The project's main homepage.
    url='https://github.com/enewe101/categorical',

    # Author details
    author='Edward Newell',
    author_email='edward.newell@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, 
		# ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
    ],

    # What does your project relate to?
    keywords='sample sampler sampling categorical multinomial discrete distribution statistics probability random',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

	install_requires=['numpy'],
)
